/* DChub - a Direct Connect Hub for Linux
 *
 * Copyright (C) 2001 Eric Prevoteau
 *
 * xmltools.h: Copyright (C) Yves BLUSSEAU
 *
 * $Id: users_xml.h,v 2.2 2003/09/28 12:16:25 blusseau Exp $
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifndef __USERS_XML_H__
#define __USERS_XML_H__

/*******************************************/
/* retreive the level for the given access */
/*************************************************************/
/* input: access: is the access to search                    */
/* output: level if the access is found, G_MAXINT otherwise  */
/*************************************************************/
int get_access_level(const char *access_name);

/*********************************************/
/* retreive the attributes of the given nick */
/*******************************************************************/
/* input: nick: is the nick to search                              */
/*        level: is a pointeur to receive the level of the user    */
/*        (can be NULL if you don't want to retreive it).          */
/*        *password: is a pointeur to receive the password of      */
/*        the user (can be NULL if you don't want to retreive it). */
/*        The password must be free()d when no more required.      */
/*        unkickable: is a pointeur to recaive a boolean that said */
/*        if the user is kickage or not (can be NULL if you don't  */
/*        want to retreive it).                                    */
/*        *comment: is a pointeur to receive comment concerning    */
/*        the user (can be NULL if you don't want to retreive it). */
/*        The comment must be free()d when no more required.       */
/*                                                                 */
/* output: TRUE if the user is found                               */
/*******************************************************************/
gboolean xml_get_user(char *nick, int *level, char **password, gboolean *unkickable, char **comment);

/******************/
/* add a new user */
/********************************/
/* output: TRUE=ok, FALSE=error */
/********************************************************************/
/* note: this function doesn't perform any check of the given value */
/********************************************************************/
gboolean xml_add_user(char *nick, int level, const char *password, gboolean unkickable, const char *comment);

/*****************/
/* delete a user */
/********************************/
/* output: TRUE=ok, FALSE=error */
/********************************************************************/
/* note: this function doesn't perform any check of the given value */
/********************************************************************/
gboolean xml_del_user(char *nicktodel);

void xml_dump_users(const char* filename);

void xml_free_users_resource();

#endif /* __USERS_XML_H__ */
